import numpy as np
import h5py


def gsf_read(filename):
    '''Read a Gwyddion Simple Field 1.0 file format
    http://gwyddion.net/documentation/user-guide-en/gsf.html

    Args:
        filename (string): the name of the output (any extension will be replaced)
    Returns:
        metadata (dict): additional metadata to be included in the file
        data (2darray): an arbitrary sized 2D array of arbitrary numeric type
    '''
    if filename.rpartition('.')[1] == '.':
        filename = filename[0:filename.rfind('.')]

    gsfFile = open(filename + '.gsf', 'rb')

    metadata = {}

    # check if header is OK
    if not (gsfFile.readline().decode('UTF-8') == 'Gwyddion Simple Field 1.0\n'):
        gsfFile.close()
        raise ValueError('File has wrong header')

    term = b'00'
    # read metadata header
    while term != b'\x00':
        line_string = gsfFile.readline().decode('UTF-8')
        metadata[line_string.rpartition('=')[0]] = line_string.rpartition('=')[2]
        term = gsfFile.read(1)

        gsfFile.seek(-1, 1)

    gsfFile.read(4 - gsfFile.tell() % 4)

    # fix known metadata types from .gsf file specs
    # first the mandatory ones...
    metadata['XRes'] = np.int(metadata['XRes'])
    metadata['YRes'] = np.int(metadata['YRes'])

    # now check for the optional ones
    if 'XReal' in metadata:
        metadata['XReal'] = np.float(metadata['XReal'])

    if 'YReal' in metadata:
        metadata['YReal'] = np.float(metadata['YReal'])

    if 'XOffset' in metadata:
        metadata['XOffset'] = np.float(metadata['XOffset'])

    if 'YOffset' in metadata:
        metadata['YOffset'] = np.float(metadata['YOffset'])

    data = np.frombuffer(gsfFile.read(), dtype='float32').reshape(metadata['YRes'], metadata['XRes'])

    gsfFile.close()

    return metadata, data


def gsf2hdf5(filename, filepath=None):
    meta, data = gsf_read(filename)
    filename = '_'.join(filename.split())
    filename = ''.join(filename.split('.')[:-1])
    filename = filename.replace('.', '-') + '.gsf'
    print(filename)
    with h5py.File(filename.split('.')[0] + ".hdf5", "w") as f:
        metadatagrp = f.create_group("metadata")
        f.create_group("process")

        name = filename.split('.')[0:-1]
        name = '_'.join(name)



        if filepath is not None:
            metadatagrp.create_dataset(filepath.split('.')[0], data=str(meta))
            datagrp = f.create_group("datasets/" + filepath.split('.')[0])
            datagrp.attrs.__setattr__('type', filepath.split('.')[-1])

        else:
            metadatagrp.create_dataset(filename.split('.')[0], data=str(meta))
            datagrp = f.create_group("datasets/" + filename.split('.')[0])
            datagrp.attrs.__setattr__('type', filename.split('.')[-1])


        datagrp.create_dataset(name, data=data)

        datagrp[name].attrs['name'] = name
        datagrp[name].attrs['shape'] = data.shape
        datagrp[name].attrs['size'] = (meta['XRes'], meta['YRes'])
        datagrp[name].attrs['offset'] = (meta['XOffset'], meta['YOffset'])
        datagrp[name].attrs['WavenumberScaling'] = float(meta['Neaspec_WavenumberScaling'])



import h5py

import numpy as np
import cv2
import time
from . import core
from . import twodim
from skimage import img_as_ubyte

#   FUNCTION l_apply_classic
# Runs m_apply multiple times successively, intended to operate on an entire process or dataset
# folder
#   INPUTS:
# filename : name of the hdf5 file where the datas are stored
# function : Custom function that you want to call
# all_input_criteria : Regex expression to describe the inputs searched for. Can be composed as a
#     list of a list of strings, with extra list parenthesis automatically generated. Eg:
#         'process*Trace1*' would pass to m_apply all files that contain 'process*Trace1*'.
#         ['process*Trace1*'] as above
#         [['process*Trace1*']] as above
#         [['process*Trace1*', 'process*Trace2*']] would pass to m_apply all files that contain
#             'process*Trace1*' and 'process*Trace2*' in a single list.
#         [['process*Trace1*'], ['process*Trace2*']] would pass to m_apply all files that contain
#             'process*Trace1*' and 'process*Trace2*' in two different lists; and thus will operate
#             differently on each of these lists.
# outputs_names (default: None): list of the names of the channels for the writting of the results.
#     By default, copies names from the first of the in_paths
# folder_names (default: None): list of the names of the folder containing results data channels.
#     By default, copies names from the first of the in_paths
# use_attrs (default: None): string, or list of strings, that are the names of attributes that will
#     be copied from in_paths, and passed into the function as a kwarg for use.
# prop_attrs (default: None): string, or list of strings, that are the names of attributes that will
#     be copied from in_paths, into each output file. If the same attribute name is in multiple
#     in_paths, the first in_path with the attribute name will be copied from.
# repeat (default: None): determines what to do if path_lists generated are of different lengths.
#     None: Default, no special action is taken, and extra entries are removed. ie, given lists
#         IJKL and AB, IJKL -> IJ.
#     'alt': The shorter lists of path names are repeated to be equal in length to the longest list.
#         ie, given IJKL and AB, AB -> ABAB
#     'block': Each entry of the shorter list of path names is repeated to be equal in length to the
#         longest list. ie, given IJKL and AB, AB -> AABB.
# **kwargs : All the non-data inputs to give to the function
#    OUTPUTS:
# null
#    TO DO:
# Can we force a None to be passed?


def l_apply_classic(filename, function, all_input_criteria, output_names=None, folder_names=None,
                    use_attrs=None, prop_attrs=None, repeat=None, **kwargs):
    print('This function is deprecated, please try to make use of l_apply instead.')
    all_in_path_list = core.path_search(filename, all_input_criteria, repeat)
    all_in_path_list = list(map(list, zip(*all_in_path_list)))
    increment_proc = True
    start_time = time.time()
    for path_num in range(len(all_in_path_list)):
        core.m_apply(filename, function, all_in_path_list[path_num], output_names=output_names,
                folder_names=folder_names, increment_proc=increment_proc,
                use_attrs=use_attrs, prop_attrs=prop_attrs, **kwargs)
        core.progress_report(path_num + 1, len(all_in_path_list), start_time, function.__name__,
                        all_in_path_list[path_num])
        increment_proc = False


def distortion_params_classic(filename, all_input_criteria, speed=2, read_offset=False,
                       cumulative=False, filterfunc=twodim.normalise):
    """
    Determine cumulative translation matrices for distortion correction and directly write it into
    an hdf5 file

    Parameters
    ----------
    filename : str
        name of hdf5 file containing data
    all_input_criteria : str
        criteria to identify paths to source files using pt.path_search. Should be
        height data to extract parameters from
    speed : int, optional
        Value between 1 and 4, which determines speed and accuracy of function. A higher number is
        faster, but assumes lower distortion and thus may be incorrect. Default value is 2.
    read_offset : bool, optional
        If set to True, attempts to read dataset for offset attributes to
        improve initial guess and thus overall accuracy (default is False).
    cumulative : bool, optional
        Determines if each image is compared to the previous image (default,
        False), or to the original image (True). Output format is identical.
    fitlerfunc : func, optional
        Function applied to image before identifying distortion params

    Returns
    -------
        None
    """

    print('This function is deprecated, please try to make use of distortion_params_ instead.')

    in_path_list = core.path_search(filename, all_input_criteria)[0]
    out_folder_locations = core.find_output_folder_location(filename, 'distortion_params',
                                                          in_path_list)
    tform21 = np.eye(2, 3, dtype=np.float32)
    cumulative_tform21 = np.eye(2, 3, dtype=np.float32)
    with h5py.File(filename, "a") as f:
        recent_offsets = []
        for i in range(len(in_path_list)):
            if i == 0:
                start_time = time.time()
            else:
                print('---')
                print('Currently reading path ' + in_path_list[i])
                i1 = f[in_path_list[0]]
                if (i > 1) and (not cumulative):
                    i1 = f[in_path_list[i-1]]
                i2 = f[in_path_list[i]]
                if filterfunc is not None:
                    i1 = filterfunc(i1)
                    i2 = filterfunc(i2)
                img1 = img_as_ubyte(i1)
                img2 = img_as_ubyte(i2)

                # try estimate offset change from attribs of img1 and img2
                if read_offset:
                    offset2 = (f[in_path_list[i]]).attrs['offset']
                    offset1 = (f[in_path_list[i - 1]]).attrs['offset']
                    scan_size = (f[in_path_list[i]]).attrs['size']
                    shape = (f[in_path_list[i]]).attrs['shape']
                    offset_px = twodim.m2px(offset2 - offset1, shape, scan_size)
                else:
                    offset_px = np.array([0, 0])
                if speed != 0 and speed != 1 and speed != 2 and speed != 3 and speed != 4:
                    print('Error: Speed should be an integer between 1 (slowest) and 4 (fastest).\
                            Speed now set to level 2.')
                    speed = 2
                if len(recent_offsets) == 0:
                    offset_guess = offset_px
                    if speed == 1:
                        warp_check_range = 16
                    elif speed == 2:
                        warp_check_range = 12
                    elif speed == 3:
                        warp_check_range = 10
                    elif speed == 4:
                        warp_check_range = 8
                elif len(recent_offsets) < 3:
                    offset_guess = offset_px + recent_offsets[-1]
                    if speed == 1:
                        warp_check_range = 12
                    elif speed == 2:
                        warp_check_range = 8
                    elif speed == 3:
                        warp_check_range = 8
                    elif speed == 4:
                        warp_check_range = 6
                else:
                    offset_guess = (offset_px + recent_offsets[2] / 2 + recent_offsets[1] / 3
                                    + recent_offsets[0] / 6)
                    # if i == 9:
                    #    offset_guess = offset_guess-np.array([20,20])
                    #    print(offset_guess)
                    if speed == 1:
                        warp_check_range = 8
                    elif speed == 2:
                        warp_check_range = 6
                    elif speed == 3:
                        warp_check_range = 4
                    elif speed == 4:
                        warp_check_range = 2
                if (offset_px[0] != 0) or (offset_px[1] != 0):
                    print('Offset found from file attributes: ' + str(offset_px))
                    warp_check_range = warp_check_range + 8
                    recent_offsets = []
                tform21 = generate_transform_xy_classic(img1, img2, tform21, offset_guess,
                                                warp_check_range, cumulative, cumulative_tform21)
                if cumulative:
                    tform21[0, 2] = tform21[0, 2] - cumulative_tform21[0, 2]
                    tform21[1, 2] = tform21[1, 2] - cumulative_tform21[1, 2]
                cumulative_tform21[0, 2] = cumulative_tform21[0, 2] + tform21[0, 2]
                cumulative_tform21[1, 2] = cumulative_tform21[1, 2] + tform21[1, 2]
                print('Scan ' + str(i) + ' Complete. Cumulative Transform Matrix:')
                print(cumulative_tform21)
                if (offset_px[0] == 0) and (offset_px[1] == 0):
                    recent_offsets.append([tform21[0, 2], tform21[1, 2]] - offset_px)
                    if len(recent_offsets) > 3:
                        recent_offsets = recent_offsets[1:]
            data = core.write_output_f(f, cumulative_tform21, out_folder_locations[i],
                                     in_path_list[i])
            core.progress_report(i + 1, len(in_path_list), start_time, 'distortion_params',
                               in_path_list[i], clear=False)


def distortion_correction_classic(filename, all_input_criteria, cropping=True):
    """
    Applies distortion correction parameters to an image. The distortion corrected data is then
    cropped to show only the common data, or expanded to show the maximum extent of all possible data.

    Parameters
    ----------
    filename : str
        Filename of hdf5 file containing data
    all_input_criteria : list
        Criteria to identify paths to source files using pt.path_search. First should
        be data to be corrected, second should be the distortion parameters.
    cropping : bool, optional
        If set to True, each dataset is cropped to show only the common area. If
        set to false, expands data shape to show all data points of all images. (default: True)

    Returns
    -------
    None
    """
    print('This function is deprecated, please try to make use of distortion_correction_ instead.')

    all_in_path_list = core.path_search(filename, all_input_criteria, repeat='block')
    in_path_list = all_in_path_list[0]
    dm_path_list = all_in_path_list[1]

    distortion_matrices = []
    with h5py.File(filename, "a") as f:
        for path in dm_path_list[:]:
            distortion_matrices.append(np.copy(f[path]))
        xoffsets = []
        yoffsets = []
        for matrix in distortion_matrices:
            xoffsets.append(np.array(matrix[0, 2]))
            yoffsets.append(np.array(matrix[1, 2]))
    offset_caps = [np.max(xoffsets), np.min(xoffsets), np.max(yoffsets), np.min(yoffsets)]

    out_folder_locations = core.find_output_folder_location(filename, 'distortion_correction',
                                                          in_path_list)

    with h5py.File(filename, "a") as f:
        start_time = time.time()
        for i in range(len(in_path_list)):
            orig_image = f[in_path_list[i]]
            if cropping:
                final_image = twodim.array_cropped(orig_image, xoffsets[i], yoffsets[i], offset_caps)
            else:
                final_image = twodim.array_expanded(orig_image, xoffsets[i], yoffsets[i], offset_caps)
            data = core.write_output_f(f, final_image, out_folder_locations[i], [in_path_list[i],
                                                                               dm_path_list[i]])
            twodim.propagate_scale_attrs(data, f[in_path_list[i]])
            core.progress_report(i + 1, len(in_path_list), start_time, 'distortion_correction',
                               in_path_list[i])


def generate_transform_xy_classic(img, img_orig, tfinit=None, offset_guess = [0,0], warp_check_range=10,
                          cumulative=False, cumulative_tform21=np.eye(2,3,dtype=np.float32)):
    """
    Determines transformation matrices in x and y coordinates

    Parameters
    ----------
    img : cv2
        Currently used image (in cv2 format) to find transformation array of
    img_orig : cv2
        Image (in cv2 format) transformation array is based off of
    tfinit : array_like or None, optional
        Base array passed into function
    offset_guess : list, optional
        Array showing initial estimate of distortion, in pixels (default: [0,0])
    warp_check_range : int, optional
        Distance (in pixels) that the function will search to find the optimal transform matrix.
        Number of iterations = (warp_check_range+1)**2. (default: 10)
    cumulative : bool, optional
        Determines if each image is compared to the previous image (default, False), or to the original image (True).
        Output format is identical.
    cumulative_tform21 : ndarray, optional
        The transformation matrix, only used if cumulative is switched to True. (default: np.eye(2,3,dtype=np.float32))

    Returns
    -------
    warp_matrix : ndarray
        Transformation matrix used to convert img_orig into img
    """

    print('This function is deprecated, please try to make use of gernerate_transform_xy_ instead.')

    # Here we generate a MOTION_EUCLIDEAN matrix by doing a
    # findTransformECC (OpenCV 3.0+ only).
    # Returns the transform matrix of the img with respect to img_orig
    warp_mode = cv2.MOTION_TRANSLATION
    if tfinit is not None:
        warp_matrix = tfinit
    else:
        warp_matrix = np.eye(2, 3, dtype=np.float32)
    number_of_iterations = 10000
    termination_eps = 1e-3
    term_flags = cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT

    if cumulative:
        offset_guess[0] = offset_guess[0] + cumulative_tform21[0, 2]
        offset_guess[1] = offset_guess[1] + cumulative_tform21[1, 2]

    criteria = (term_flags, number_of_iterations, termination_eps)

    diff = np.Inf
    offset1 = 0
    offset2 = 0
    for i in range(-warp_check_range // 2, (warp_check_range // 2) + 1):
        for j in range(-warp_check_range // 2, (warp_check_range // 2) + 1):
            warp_matrix[0, 2] = 2 * i + offset_guess[0]
            warp_matrix[1, 2] = 2 * j + offset_guess[1]
            try:
                (cc, tform21) = cv2.findTransformECC(img_orig, img, warp_matrix, warp_mode,
                                                     criteria)
                img_test = cv2.warpAffine(img, tform21, (512, 512), flags=cv2.INTER_LINEAR +
                                                                          cv2.WARP_INVERSE_MAP)
                currDiff = np.sum(np.square(img_test[150:-150, 150:-150]
                                            - img_orig[150:-150, 150:-150]))
                if currDiff < diff:
                    diff = currDiff
                    offset1 = tform21[0, 2]
                    offset2 = tform21[1, 2]
            except:
                pass
            warp_matrix[0, 2] = offset1
            warp_matrix[1, 2] = offset2
    return warp_matrix




def sample_fraction(array, shape_fraction=0.1, start_shape=None):
    """
    Randomly samples a contiguous block and small fraction of a larger array. Works by taking a
    small point (or smaller shape) and randomly expanding in all directions. To ensure uniformity,
    a 'supershape' is constructed, which takes the initial array and adds 5 pixels on each
    direction to reduce edge effects during sampling.

    Parameters
    ----------
    array : 2d array
        array to be sampled
    shape_fraction : int or float, optional
        the fraction of the array that is actually sampled. eg., by
        default, an area of 0.1 will return a block of area 10% of the overall array
    start_shape : 2d array, optional
        A shape used to start the sample. Allows sample_fraction to feed into
        itself and generate larger samples from smaller ones.

    Returns
    ------
    curr_shape :
        the sample generated from the main array
    """
    basic_array = np.zeros_like(array)
    total_shapes = 300000
    target_shape_size = shape_fraction * np.sum(array)

    # Create Coordinate List:
    coord_list = []
    coord_list_index_array = np.zeros_like(basic_array)
    index = 0
    for i in range(np.shape(basic_array)[0]):
        for j in range(np.shape(basic_array)[1]):
            coord_list.append([i, j])
            coord_list_index_array[i, j] = index
            index = index + 1
    neighbour_array = np.zeros_like(basic_array)
    neighbour_scaling = np.zeros_like(basic_array) + 4

    if start_shape is None:
        # Pick First Point
        np.random.seed()
        init_i, init_j = coord_list[int(np.random.choice(coord_list_index_array.flatten()))]
        new_i = init_i
        new_j = init_j
        curr_shape = np.zeros_like(basic_array)
        curr_shape[init_i, init_j] = 1
    else:
        start_shape = np.copy(start_shape)
        new_i = None
        new_j = None
        curr_shape = start_shape
        expanded_point = np.zeros_like(basic_array)
        for i in range(np.shape(basic_array)[0]):
            for j in range(np.shape(basic_array)[1]):
                if curr_shape[i, j] != 0:
                    if i != 0:
                        if i == 1:
                            expanded_point[i - 1, j] = expanded_point[i - 1, j] + 2
                        else:
                            expanded_point[i - 1, j] = expanded_point[i - 1, j] + 2
                    if j != 0:
                        if j == 1:
                            expanded_point[i, j - 1] = expanded_point[i, j - 1] + 2
                        else:
                            expanded_point[i, j - 1] = expanded_point[i, j - 1] + 2
                    if i != np.shape(curr_shape)[0] - 1:
                        if i == np.shape(curr_shape)[0] - 2:
                            expanded_point[i + 1, j] = expanded_point[i + 1, j] + 2
                        else:
                            expanded_point[i + 1, j] = expanded_point[i + 1, j] + 2
                    if j != np.shape(curr_shape)[1] - 1:
                        if j == np.shape(curr_shape)[1] - 2:
                            expanded_point[i, j + 1] = expanded_point[i, j + 1] + 2
                        else:
                            expanded_point[i, j + 1] = expanded_point[i, j + 1] + 2

    curr_shape_size = np.sum(curr_shape * array)
    while curr_shape_size < target_shape_size:
        # Find possible growths from that point
        last_point = np.zeros_like(basic_array)
        last_point[new_i, new_j] = 1
        # Use mirror boundary conditions
        if new_i is not None:
            expanded_point = np.copy(last_point)
            if new_i > 1:
                expanded_point[new_i - 1, new_j] = 1
            elif new_i == 1:
                expanded_point[new_i - 1, new_j] = 2
            if new_j > 1:
                expanded_point[new_i, new_j - 1] = 1
            elif new_j == 1:
                expanded_point[new_i, new_j - 1] = 2
            if new_i < np.shape(basic_array)[0] - 2:
                expanded_point[new_i + 1, new_j] = 1
            elif new_i == np.shape(basic_array)[0] - 2:
                expanded_point[new_i + 1, new_j] = 2
            if new_j < np.shape(basic_array)[1] - 2:
                expanded_point[new_i, new_j + 1] = 1
            elif new_j == np.shape(basic_array)[1] - 2:
                expanded_point[new_i, new_j + 1] = 2
        neighbour_array = neighbour_array + (expanded_point / neighbour_scaling)
        neighbour_array = neighbour_array * (1 - curr_shape)

        # Grow from one of the growth coords
        thresh = 0.75
        if np.any(neighbour_array >= thresh):
            # Grow if too surrounded
            total_to_change = np.sum(neighbour_array >= thresh)
            # print(total_to_change)
            if total_to_change > 1:
                growth_locations = np.where(neighbour_array >= thresh)
                rand_num = np.random.randint(0, len(growth_locations[0]))
                new_i = growth_locations[0][rand_num]
                new_j = growth_locations[1][rand_num]
            else:
                new_coords = np.where(neighbour_array >= thresh)
                new_i = new_coords[0]
                new_j = new_coords[1]
        else:
            # Randomly grow:
            probability_array = np.copy(neighbour_array)
            normalise_factor = np.sum(probability_array)
            new_i, new_j = coord_list[int(np.random.choice(coord_list_index_array.flatten(),
                                                           p=(probability_array / normalise_factor).flatten()))]
        curr_shape[new_i, new_j] = 1
        curr_shape_size = np.sum(curr_shape * array)
        # Should make it only consider largest area?
    return curr_shape



def all_sample_fractions(array, iterations=100, fractions=[0.1,0.15,0.2,0.25],
                         compression=[50,50], background=np.nan):
    """
    Generates a 4D array of several samples; the first axis allows choice of the fraction of each
    sampling; the second axis allows for each iteration of this fraction. The remaining two axes
    are the 2D array of each individual sample.

    Parameters
    ----------
    array : 2d array
        the array to be sampled
    iterations : int or float, optional
        number of samples to be taken of for each fraction
    fractions : list, optional
        fractions to be sampled of the array
    compression : list, optional
        the size to which the array is compressed to during sampling
    background : int or float, optional
        values of background areas that may be removed during compression

    Returns
    ------
    sample_fractions_all_fractions :
        4D array describing all samples extracted
    """
    # Compress array and generate "supershape" template from which subshapes are drawn
    if np.isnan(background):
        bool_array = ~np.isnan(array)
    else:
        bool_array = ~(array == background)
    cropped_array = crop(bool_array)
    expanded_array = uncrop_to_multiple(cropped_array, compression)
    compressed_array = compress_to_shape(expanded_array, compression)
    array_supershape = uncrop_to_multiple(compressed_array, [compression[0]+10,
                                                             compression[1]+10])

    # Generate first generation of all shapes
    if type(fractions) != list:
        fractions = [fractions]
    sample_fractions_all_fractions = []
    sample_fractions_one_fraction = []
    for sample_count in range(iterations):
        shape = sample_fraction(array_supershape, fractions[0]).astype(bool)
        sample_fractions_one_fraction.append(shape)

    # If multiple fractions are provides, generate successive generations
    if len(fractions) > 1:
        for i in range(1, len(fractions)):
            sample_fractions_all_fractions.append(sample_fractions_one_fraction)
            sample_fractions_one_fraction = []
            for sample_count in range(iterations):
                shape = sample_fraction(array_supershape, fractions[i],
                                        sample_fractions_all_fractions[i - 1]
                                        [sample_count]).astype(bool)
                sample_fractions_one_fraction.append(shape)
        sample_fractions_all_fractions.append(sample_fractions_one_fraction)
    else:
        sample_fractions_all_fractions = sample_fractions_one_fraction

    sample_fractions_all_fractions = np.array(sample_fractions_all_fractions)
    # axes = [(fraction_num)][sample_num][y][x]
    return sample_fractions_all_fractions


def interpolated_features(switchmap):
    """
    Creates isolines from a switchmap, then interpolates it

    Parameters
    ----------
    switchmap : 2d array
        the switchmap used as the base for key features

    Returns
    ------
    interpolation :
        interpolated features
    """
    isolines = find_isolines(switchmap)

    isoline_y = []
    isoline_x = []
    isoline_z = []
    for i in range(np.shape(isolines)[0]):
        for j in range(np.shape(isolines)[1]):
            if isolines[i, j] != 0:
                isoline_x.append(j)
                isoline_y.append(i)
                isoline_z.append(isolines[i, j])
    grid_x, grid_y = np.mgrid[0:np.shape(isolines)[0]:1, 0:np.shape(isolines)[1]:1]
    interpolation = interpolate.griddata(np.array([isoline_y, isoline_x]).T, np.array(isoline_z),
                                         (grid_x, grid_y), method='linear', fill_value=np.nan)
    return interpolation


import h5py
import importlib
from hystorian import processing
import os
import ast
import sys
import numpy as np

def compress_hdf5(file, error_threshold=0, bypass_verification=False, dependencies=[]):
    identical = True
    if not bypass_verification:
        print('Please understand that this function WILL remove processed datas from your hdf5.\n '
              'Normally they should be rebuildable using existing functions. But no guarantee is given. \n '
              'Please write "I understand the risk"')
        input_value = input()
        if input_value != 'I understand the risk':
            return None

    with h5py.File(file, 'r+') as f:
        processes = f['process']
        processing_lst = processes.keys()
        for k in list(processing_lst)[::-1]:
            fname = list(processes[k].keys())[0]
            outputs = list(processes[k][fname])

            module = importlib.import_module('.'.join(
                processes[k][fname][outputs[0]].attrs['operation name'].split('.')[:-1]))
            if processes[k][fname][outputs[0]].attrs['operation name'].split('.')[0] == '__main__':
                use_exec = True
            # if module
            else:
                use_exec = False
                func = getattr(module,
                           processes[k][fname][outputs[0]].attrs['operation name'].split('.')[-1])

            kwargs = {}
            if processes[k][fname].attrs.get('compressed') is None:

                for key in processes[k][fname][outputs[0]].attrs.keys():
                    if key.split('_')[0] == 'kwargs':
                        short_key = '_'.join(key.split('_')[1:])
                        kwargs[short_key] = processes[k][fname][outputs[0]].attrs[key]

                inputs = []
                for source in processes[k][fname][outputs[0]].attrs['source']:
                    inputs.append(f[source][()])
                # return inputs
                #input()
                #try:
                if use_exec:
                    fct_code = processes[k][fname][outputs[0]].attrs['function code']

                    f = open("0123456789_tmp_pyfile.py", "w")
                    for d in dependencies:
                        if len(d) == 2:
                            f.write('import ' + d[0] + ' as ' + d[1] +'\n')
                        elif len(d) == 1:
                            f.write('import ' + d + '\n')
                    f.write(fct_code)
                    f.close()

                    import tmp_pyfile

                    func = getattr(tmp_pyfile,
                                   processes[k][fname][outputs[0]].attrs['operation name'].split('.')[-1])

                    results = func(*inputs, **kwargs)
                    os.remove("0123456789_tmp_pyfile.py")
                    processes[k][fname][outputs[0]].attrs['dependencies'] = dependencies

                else:
                    results = func(*inputs, **kwargs)

                if type(results) != tuple:
                    results = tuple([results])
                for i in range(len(outputs)):
                    if type(results[i]) == dict:
                        if 'hdf5_dict' in results[i]:
                            result_data = results[i]['data']
                        else:
                            result_data = results[i]
                    else:
                        result_data = results[i]
                    if (result_data - processes[k][fname][outputs[i]][()] < error_threshold).all():
                        print(func.__name__ + ': Result is not identical, not compressing')
                        identical = False
                        break
                #except Exception as e:
                #    print('ERROR: ' + e.__doc__)
                #    identical = False

                if identical:
                    print(func.__name__ + ': Result is identical, compressing')
                    for i in range(len(outputs)):
                        tmpattrs = {}
                        for k2 in processes[k][fname][outputs[i]].attrs.keys():
                            tmpattrs[k2] = processes[k][fname][outputs[i]].attrs[k2]

                        del (processes[k][fname][outputs[i]])
                        processes[k][fname].create_dataset(outputs[i], (1,), dtype=int)
                        for k3 in tmpattrs.keys():
                            val = tmpattrs[k3]
                            processes[k][fname][outputs[i]].attrs.__setitem__(k3, val)
                    processes[k][fname].attrs.__setitem__('compressed', True)
                else:
                    processes[k][fname].attrs.__setitem__('compressed', False)

    processing.core.deallocate_hdf5_memory(file, verify=False)

def decompress_hdf5(file):
    with h5py.File(file, 'r+') as f:
        processes = f['process']
        processing_lst = processes.keys()

        for k in list(processing_lst)[::-1]:
            fname = list(processes[k].keys())[0]
            if processes[k][fname].attrs.get('compressed'):
                outputs = list(processes[k][fname])
                module = importlib.import_module('.'.join(
                    processes[k][fname][outputs[0]].attrs['operation name'].split('.')[:-1]))
                if processes[k][fname][outputs[0]].attrs['operation name'].split('.')[0] == '__main__':
                    use_exec = True

                kwargs = {}
                for key in processes[k][fname][outputs[0]].attrs.keys():
                    if k.split('_')[0] == 'kwargs':
                        short_key = '_'.join(key.split('_')[1:])
                        kwargs[short_key] = processes[k][fname][outputs[0]].attrs[k]

                inputs = []
                for source in processes[k][fname][outputs[0]].attrs['source']:
                    inputs.append(f[source][()])

                if use_exec:
                    fct_code = processes[k][fname][outputs[0]].attrs['function code']

                    f = open("tmp_pyfile.py", "a")
                    for d in processes[k][fname][outputs[0]].attrs['dependencies']:
                        if len(d) == 2:
                            f.write('import ' + d[0] + ' as ' + d[1] + '\n')
                        elif len(d) == 1:
                            f.write('import ' + d + '\n')
                    f.write(fct_code)
                    f.close()

                    import tmp_pyfile

                    func = getattr(tmp_pyfile,
                                   processes[k][fname][outputs[0]].attrs['operation name'].split('.')[-1])

                    results = func(*inputs, **kwargs)
                    os.remove("tmp_pyfile.py")
                else:
                    func = getattr(module,
                                   processes[k][fname][outputs[0]].attrs['operation name'].split('.')[-1])


                    results = func(*inputs, **kwargs)
                if type(results) != tuple:
                    results = tuple([results])

                for i in range(len(outputs)):
                    tmpattrs = {}
                    for k2 in processes[k][fname][outputs[i]].attrs.keys():
                        tmpattrs[k2] = processes[k][fname][outputs[i]].attrs[k2]

                    del (processes[k][fname][outputs[i]])
                    if processes[k][fname].attrs.__contains__('compressed'):
                        processes[k][fname].attrs.__delitem__('compressed')
                    processes[k][fname].create_dataset(outputs[i], data=results[i])
                    for k3 in tmpattrs.keys():
                        val = tmpattrs[k3]
                        processes[k][fname][outputs[i]].attrs.__setitem__(k3, val)


import os
import numpy as np
import inspect
import struct
import h5py

DEBUG_PRINT_INFO = True
DEBUG_PRINT_WARNING = True
DEBUG_PRINT_ERROR = True


def string_scinot_to_float(string):
    split_string = string.split('E')
    float_ = float(split_string[0])*10**float(split_string[1])
    return float_


def debug_print_info(s):
    if not DEBUG_PRINT_INFO:
        return
    print("[" + str(inspect.stack()[1][3]) + "] Info: " + str(s))


def debug_print_warning(s):
    if not DEBUG_PRINT_INFO:
        return
    print("[" + str(inspect.stack()[1][3]) + "] Warning: " + str(s))


def debug_print_error(s):
    if not DEBUG_PRINT_INFO:
        return
    print("[" + str(inspect.stack()[1][3]) + "] Error: " + str(s))


def parse_header(header_data):
    header_dict = {}

    current_item = ""
    current_value = []
    for line in header_data.decode("ascii").split("\n"):
        if line.startswith(":"):
            if current_item != "":
                if len(current_value) == 1:
                    current_value = current_value[0]
                header_dict[current_item] = current_value
                current_value = []
            current_item = line
            if line == ":SCANIT_END:":
                continue
        else:
            if current_item != "":
                current_value.append(line)
    return header_dict


def parse_channel_info(header):
    if ":DATA_INFO:" not in header:
        debug_print_error("Cannot find :DATA_INFO: in header!!")
        return -1

    '''if ":Scan>channels:" not in header:
        debug_print_error("Cannot find :Scan>channels: in header!!")
        return -1
    '''
    try:
        fullnames = header[":Scan>channels:"].split(";")
    except ValueError:
        fullnames = []

    fullnameindex = 0
    channels = []

    for c in header[":DATA_INFO:"]:
        chan = {}

        if len(c.strip()) == 0:
            continue

        c_num = c.split("\t")[1]
        c_name = c.split("\t")[2]
        c_unit = c.split("\t")[3]
        c_direction = c.split("\t")[4]
        c_calibration = c.split("\t")[5]
        c_offset = c.split("\t")[6]

        if c_num == "Channel":
            continue
        try:
            chan["Index"] = fullnames.index(c_name.replace("_", " ") + " (" + str(c_unit) + ")")
        except:
            chan["Index"] = -1

        chan["Channel"] = int(c_num)
        chan["Name"] = str(c_name)
        try:
            chan["Full Name"] = fullnames[chan["Index"]]
        except:
            chan["Full Name"] = chan["Name"]
        chan["Unit"] = str(c_unit)
        chan["Direction"] = str(c_direction)
        chan["Calibration"] = float(c_calibration)
        chan["Offset"] = float(c_offset)

        channels.append(chan)
    return channels


def get_channels(channel_info, orientation, fd, data_start_pos, pixels_x, pixels_y):
    channel_data = {}
    fd.seek(data_start_pos + 1)
    for c in channel_info:
        if c["Direction"] == "both":
            trace_data = np.reshape(
                np.array(struct.unpack(">" + str(pixels_x * pixels_y) + "f", fd.read(4 * pixels_x * pixels_y))),
                (pixels_x, pixels_y))
            if orientation == 'up':
                trace_data = np.flip(trace_data, axis=0)
            channel_data[c["Full Name"] + " - Trace"] = trace_data

            retrace_data = np.reshape(
                np.array(struct.unpack(">" + str(pixels_x * pixels_y) + "f", fd.read(4 * pixels_x * pixels_y))),
                (pixels_x, pixels_y))
            if orientation == 'up':
                retrace_data = np.flip(retrace_data, axis=0)
            channel_data[c["Full Name"] + " - Retrace"] = np.flip(retrace_data, axis=1)
        else:
            debug_print_info("Scan direction is not 'both'. Assuming single trace/retrace acquisition...")
            trace_data = np.reshape(
                np.array(struct.unpack(">" + str(pixels_x * pixels_y) + "f", fd.read(4 * pixels_x * pixels_y))),
                (pixels_x, pixels_y))
            channel_data[c["Full Name"]] = trace_data
    return channel_data


def load_sxm(filename):
    header_start_token = ':NANONIS_VERSION:'
    header_end_token = ':SCANIT_END:\n\n\n'

    fd = open(filename, 'rb')

    dat = fd.read(1000000)

    if dat.find(header_start_token.encode()) != 0:
        debug_print_error("Cannot find :NANONIS_VERSION: header... Is this a Nanonis file??")

    header_end_pos = dat.find(header_end_token.encode())
    data_start_pos = header_end_pos + len(header_end_token) + 1
    data_end_pos = fd.seek(0, os.SEEK_END)
    fd.seek(0)
    header_bytes = fd.read(header_end_pos + len(header_end_token) + 1)

    header = parse_header(header_bytes)
    channel_info = parse_channel_info(header)
    # print(channel_info)
    # print(header.keys())
    orientation = header[':SCAN_DIR:']
    if ":Scan>pixels/line:" in header.keys():
        channel_data = get_channels(channel_info, orientation, fd, data_start_pos, int(header[":Scan>lines:"]),
                                    int(header[":Scan>pixels/line:"]))
    else:
        px = header[':SCAN_PIXELS:'].split()[0]
        ln = header[':SCAN_PIXELS:'].split()[1]
        channel_data = get_channels(channel_info, orientation, fd, data_start_pos, int(ln),
                                    int(px))

    return [header, channel_info, channel_data]


def sxm2hdf5(filename, filepath=None):
    header, channel_info, channel_data = load_sxm(filename)

    with h5py.File(filename.split('.')[0] + ".hdf5", "w") as f:
        metadatagrp = f.create_group("metadata")
        f.create_group("process")


        if filepath is not None:
            metadatagrp.create_dataset(filepath.split('.')[0], data=str(header))
            datagrp = f.create_group("datasets/" + filepath.split('.')[0])
            datagrp.attrs.__setattr__('type', filepath.split('.')[-1])

        else:
            metadatagrp.create_dataset(filename.split('.')[0], data=str(header))
            datagrp = f.create_group("datasets/" + filename.split('.')[0])
            datagrp.attrs.__setattr__('type', filename.split('.')[-1])



        for indx, key in enumerate(channel_data.keys()):
            datagrp.create_dataset(key, data=channel_data[key])
            datagrp[key].attrs['name'] = key
            datagrp[key].attrs['shape'] = (int(header[':Scan>pixels/line:']), int(header[':Scan>lines:']))
            offsets = header[':SCAN_OFFSET:'].split(' ')
            while '' in offsets:
                offsets.remove('')
            datagrp[key].attrs['offset'] = (string_scinot_to_float(offsets[0]), string_scinot_to_float(offsets[1]))
            sizes = header[':SCAN_RANGE:'].split(' ')
            while '' in sizes:
                sizes.remove('')
            datagrp[key].attrs['size'] = (string_scinot_to_float(sizes[0]), string_scinot_to_float(sizes[1]))
            
            for info in channel_info:
                if info['Name'] == key.split(' - ')[0]:
                    for k2 in info.keys():
                        datagrp[key].attrs[k2] = info[k2]
                continue



from . import core
import numpy as np
import h5py
import matplotlib.pyplot as plt

def extract_hist(*input_data, bias):
    """
    Split the SSPFM curve into on and off signal.
    Function was rewritten to work with m_apply, and to take account a bug into Cypher machine,
    leading to pulse not all being the same size.

    Parameters
    ----------
    input_data : array-like
        data which needs to be splitted
    bias : array-like
        data corresponding to the bias channel

    Returns
    -------
    output : array-like
        return two grid with the data splitted into "on/off" according to the bias shape

    """
    idx_zeroes = np.where(bias[0, 0, :] == 0)[0]
    idx_Nonzeroes = np.where(bias[0, 0, :] != 0)[0]

    v1 = idx_zeroes[:-1]
    v2 = np.roll(idx_zeroes, 1)[:-1]
    list_change_zeroes = list(np.where((v1 - v2) != 1)[0])

    idx_start = idx_zeroes[list_change_zeroes]

    v1 = idx_Nonzeroes[:-1]
    v2 = np.roll(idx_Nonzeroes, 1)[:-1]
    list_change_nonzeroes = list(np.where((v1 - v2) != 1)[0])
    idx_end = idx_Nonzeroes[list_change_nonzeroes]

    data_k_on = np.ndarray((np.shape(bias)[0], np.shape(bias)[1], len(idx_end) - 1))
    data_k_off = np.ndarray((np.shape(bias)[0], np.shape(bias)[1], len(idx_end) - 1))

    input_data = input_data[0]
    for i in range(len(idx_end) - 1):
        data_k_on[:, :, i] = np.median(input_data[:, :, idx_end[i]:idx_start[i + 1]], axis=2)
        data_k_off[:, :, i] = np.median(input_data[:, :, idx_start[i + 1]:idx_end[i + 1]], axis=2)

    output = [data_k_on, data_k_off]
    output = tuple(output)
    return output

def get_phase_unwrapping_shift_(phase, phase_step=1):
    '''
    Finds the smallest shift that minimizes the phase jump in a phase series
    
    Parameters
    ----------
    phase : a 1D array of consecutive phase measurements
    phase_step : the algorithm will try shifts every phase_step. Increase this to speed up execution.
    
    Returns
    -------
    The phase shift. (phase + shift) % 360 will have the smallest jumps between consecutive phase points.
    '''
    phase_step = 90
    jumps = []
    for shift in range(0, 360, phase_step):
        y = (phase + shift) % 360
        #what is the biggest phase jump?
        max_phase_jump = np.max(np.abs(np.diff(y)))
        jumps.append(max_phase_jump)
        
    return phase_step * np.argmin(np.array(jumps))

def calc_hyst_params(bias, phase):
    """
    Calculate hysteresis parameters from bias and phase channels.
    Used in PFM_params_map. Would recommend using the function instead.

    Parameters
    ----------
    bias : array-like
        array containing the bias acquired during an SSPFM
    phase : array-like
        array containing the phase acquired during an SSPFM

    Returns
    -------
        list
        list where the elements correspond to :
            the positive coercive bias
            the negative coercive bias
            The size of the phase jump (left side)
            The size of the phase jump (right side)
    """
    biasdiff = np.diff(bias)
    up = np.sort(np.unique(np.hstack((np.where(biasdiff > 0)[0], np.where(biasdiff > 0)[0] + 1))))
    dn = np.sort(np.unique(np.hstack((np.where(biasdiff < 0)[0], np.where(biasdiff < 0)[0] + 1))))
    phase_shift = get_phase_unwrapping_shift_(phase)

    # UP leg calculations
    if up.size == 0:
        step_left_up = np.nan
        step_right_up = np.nan
        coercive_volt_up = np.nan
    else:
        x = np.array(bias[up])
        y = (np.array(phase[up])+phase_shift) % 360
        step_left_up = np.median(y[np.where(x == np.min(x))[0]])
        step_right_up = np.median(y[np.where(x == np.max(x))[0]])
    
        avg_x = []
        avg_y = []
        for v in np.unique(x):
            avg_x.append(v)
            avg_y.append(np.mean(y[np.where(x == v)[0]]))
    
        my_x = np.array(avg_x)[1:]
        my_y = np.abs(np.diff(avg_y))
    
        coercive_volt_up = my_x[np.nanargmax(my_y)]

    # DOWN leg calculations
    if dn.size == 0:
        step_left_dn = np.nan
        step_right_dn = np.nan
        coercive_volt_dn = np.nan
    else:
        x = np.array(bias[dn])
        y = (np.array(phase[dn])+phase_shift) % 360
        step_left_dn = np.median(y[np.where(x == np.min(x))[0]])
        step_right_dn = np.median(y[np.where(x == np.max(x))[0]])
    
        avg_x = []
        avg_y = []
        for v in np.unique(x):
            avg_x.append(v)
            avg_y.append(np.mean(y[np.where(x == v)[0]]))
    
        my_x = np.array(avg_x)[1:]
        my_y = np.abs(np.diff(avg_y))
    
        coercive_volt_dn = my_x[np.nanargmax(my_y)]

    return [coercive_volt_up, coercive_volt_dn, np.nanmean([step_left_dn, step_left_up]),
            np.nanmean([step_right_dn, step_right_up])]


def PFM_params_map(bias, phase):
    """
    Calculate physically relevant hysteresis parameters from bias and phase channels.

    Parameters
    ----------
    bias : array-like
        array containing the bias acquired during an SSPFM
    phase : array-like
        array containing the phase acquired during an SSPFM

    Returns
    -------
        list
        list where the elements correspond to :
            the positive coercive bias
            the negative coercive bias
            The size of the phase jump (left side)
            The size of the phase jump (right side)
            the imprint of the switching loop
            the phase shift of the switching loop
    """
    x, y, z = np.shape(phase)
    coerc_pos = np.zeros((x, y), dtype=float)
    coerc_neg = np.zeros((x, y), dtype=float)
    step_left = np.zeros((x, y), dtype=float)
    step_right = np.zeros((x, y), dtype=float)
    imprint = np.zeros((x, y), dtype=float)
    phase_shift = np.zeros((x, y), dtype=float)
    for xi in range(x):
        for yi in range(y):
            hyst_matrix = calc_hyst_params(bias[xi, yi, :], phase[xi, yi, :])
            coerc_pos[xi, yi] = hyst_matrix[0]
            coerc_neg[xi, yi] = hyst_matrix[1]
            step_left[xi, yi] = hyst_matrix[2]
            step_right[xi, yi] = hyst_matrix[3]
            imprint[xi, yi] = (hyst_matrix[0] + hyst_matrix[1]) / 2.0
            phase_shift[xi, yi] = (hyst_matrix[3] - hyst_matrix[2])

    return coerc_pos, coerc_neg, step_left, step_right, imprint, phase_shift


def gauss_area(x, y):
    """
    Determine the area created by the polygon formed by x,y using the Gauss's area formula (also
    called shoelace formula)

    Parameters
    ----------
    x  : array_like
        values along the x axis
    y : array_like
        values along the x axis

    Returns
    -------
        float
            value corresponding at the encompassed area
    """

    area = 0.0
    for i in range(len(x)):
        x1 = x[i]
        y1 = y[i]

        if i < len(x) - 1:
            x2 = x[i + 1]
            y2 = y[i + 1]
        else:
            x2 = x[0]
            y2 = y[0]

        area = area + x1 * y2 - x2 * y1
    return np.abs(area / 2.0)


def clean_loop(bias, phase, amp, threshold=None, debug=False):
    '''
    Used to determine if a SSPFM loop is good or not by calculating the area encompassed by the
    hysteresis curve and comparing it to a threshold

    Parameters
    ----------
    bias: nd-array
        3d-array containing the bias applied to each point of the grid
    phase: nd-array
        3d-array containing the phase applied to each point of the grid
    amp: nd-array
        3d-array containing the amplitude applied to each point of the grid
    threshold: int or float, optional
        minimal value of the loop area to be considered a good loop (default: None)
        if set to None will use threshold = np.mean(area_grid_full) - 2 * np.std(area_grid_full)
    debug: bool, optional
        if set to True will use plt.imshow() to plot the loops in red for bad and blue for good.
        (default: False)

    Returns
    -------
        list_bias: array-like
            list of the bias corresponding to good loops
        list_phase: array-like
            list of the phase corresponding to good loops
        list_amp: array-like
            list of the amplitudes corresponding to good loops
        mask: nd-array
            a 2D mask where a 1 correspond to a good loop and 0 to a bad loop, can be used to mask
            the input data.
    '''
    list_bias = []
    list_phase = []
    list_amp = []
    list_bias_bad = []
    list_phase_bad = []
    mask = np.ndarray((np.shape(bias)[0], np.shape(bias)[1]))

    if threshold is None:
        area_grid_full = np.ndarray((np.shape(bias)[0], np.shape(bias)[1]))
        for xi in range(np.shape(bias)[0]):
            for yi in range(np.shape(bias)[1]):
                area_grid_full[xi, yi] = gauss_area(bias[xi, yi, :], phase[xi, yi, :])
            threshold = np.mean(area_grid_full) - 2 * np.std(area_grid_full)

    for xi in range(np.shape(bias)[0]):
        for yi in range(np.shape(bias)[1]):
            if gauss_area(bias[xi, yi, :], phase[xi, yi, :]) > threshold:
                list_bias.append(bias[xi, yi, :])
                list_phase.append(phase[xi, yi, :])
                list_amp.append(amp[xi, yi, :])
                mask[xi, yi] = True
            else:
                mask[xi, yi] = False
                if debug:
                    list_bias_bad.append(bias[xi, yi, :])
                    list_phase_bad.append(phase[xi, yi, :])

    if debug:
        plt.figure(figsize=(10, 5))

        if threshold is None:
            plt.plot(np.ravel(area_grid_full))

        else:
            area_grid_full = np.ndarray((np.shape(bias)[0], np.shape(bias)[1]))
            for xi in range(np.shape(bias)[0]):
                for yi in range(np.shape(bias)[1]):
                    area_grid_full[xi, yi] = gauss_area(bias[xi, yi, :], phase[xi, yi, :])
            plt.plot(np.ravel(area_grid_full))

        plt.axhline(threshold, c='r')

    return list_bias, list_phase, list_amp, mask


def negative_(filename, data_folder='datasets', selection=None, criteria=0):
    """
    Processes an array determine negatives of all values
    Trivial sample function to show how to use proc_tools

    Parameters
    ----------
    filename : str
        name of the hdf5 file in which the data you want to apply the function to is contained
    data_folder : str, optional
        path in the hdf5 where the data is contained (default: 'datasets')
    selection : str, optional
        determines the name of folders or files to be used. Can be None which selects all data
        (default), a string, or a list of strings
    criteria : int
        determines category of files to search
    Returns
    -------
        None
    """
    # Trivial sample function to show how to use proc_tools
    # Processes an array determine negatives of all values
    in_path_list = core.path_inputs(filename, data_folder, selection, criteria)
    out_folder_locations = core.find_output_folder_location(filename, 'negative', in_path_list)
    with h5py.File(filename, "a") as f:
        for i in range(len(in_path_list)):
            neg = -np.array(f[in_path_list[i]])
            core.write_output_f(f, neg, out_folder_locations[i], in_path_list[i])



import numpy as np
import struct
import zlib
import os
import h5py
import gc


# ==========================================
# ARDF conversion
def debug_print(dstr):
    if not dstr.startswith("Info"):
        print(dstr)


def ARDF_get_next_element(fd):
    try:
        header = fd.read(0x8)
    except:
        debug_print("Error: ARDF_get_next_element - cannot read header!")
        return -1

    crc32, size = ARDF_parse_header(header)

    data_size = size - 0x8

    if data_size <= 0:
        debug_print("Error: ARDF_get_next_element - not a valid element! (assert size > 0)")
        return -1

    try:
        data = fd.read(size - 0x8)
    except:
        debug_print("Error: ARDF_get_next_element - cannot read data!")
        return -1

    if ARDF_check_header(header, data, crc32) != 0:
        debug_print("Error: ARDF_get_next_element - Error validating header CRC32!")
        return -1

    return [size, data]


def ARDF_parse_header(header):
    try:
        crc32 = struct.unpack("<L", header[0:4])[0]
        size = struct.unpack("<L", header[4:8])[0]
        debug_print("Info: ARDF_parse_header - CRC32=" + str(crc32) + " SIZE=" + str(size))
    except:
        debug_print("Error: ARDF_parse_header - cannot parse!")
        print(header)
        return -1

    return [crc32, size]


def ARDF_check_header(header, data, crc32):
    if zlib.crc32(header[4:8] + data) % (1 << 32) != crc32:
        debug_print("Error: ARDF_check_header - header CRC32 does not match!")
        return -1

    return 0


def ARDF_get_FTOC_entry(fd, ftoc_entry_size):
    try:
        header = fd.read(ftoc_entry_size)
    except:
        debug_print("Error: ARDF_get_FTOC_entry - cannot read header!")
        return -1

    crc32, size = ARDF_parse_header(header)

    if crc32 == 0 or size == 0:
        debug_print("Info: ARDF_get_FTOC_entry - Not a valid entry, maybe empty...")
        return -1

    if zlib.crc32(header[4:]) % (1 << 32) != crc32:
        debug_print("Error: ARDF_get_FTOC_entry - header CRC32 does not match!")
        return -1

    crc32 = struct.unpack("<L", header[0:4])[0]
    size = struct.unpack("<L", header[4:8])[0]
    magic = struct.unpack("<4s", header[8:12])[0].decode("ascii")
    unk1 = struct.unpack("<L", header[12:16])[0]
    offset = struct.unpack("<L", header[16:20])[0]
    unk2 = struct.unpack("<L", header[20:24])[0]

    return [magic, offset, size, [crc32, unk1, unk2]]


def ARDF_get_TTOC_entry(fd, ttoc_entry_size):
    try:
        header = fd.read(ttoc_entry_size)
    except:
        debug_print("Error: ARDF_get_TTOC_entry - cannot read header!")
        return -1

    crc32, size = ARDF_parse_header(header)

    if crc32 == 0 or size == 0:
        debug_print("Info: ARDF_get_TTOC_entry - Not a valid entry, maybe empty...")
        return -1

    if zlib.crc32(header[4:]) % (1 << 32) != crc32:
        debug_print("Error: ARDF_get_TTOC_entry - header CRC32 does not match!")
        return -1

    crc32 = struct.unpack("<L", header[0:4])[0]
    size = struct.unpack("<L", header[4:8])[0]
    magic = struct.unpack("<4s", header[8:12])[0].decode("ascii")
    unk1 = struct.unpack("<L", header[12:16])[0]
    unk2 = struct.unpack("<L", header[16:20])[0]
    unk3 = struct.unpack("<L", header[20:24])[0]
    offset = struct.unpack("<L", header[24:28])[0]
    unk4 = struct.unpack("<L", header[28:32])[0]

    return [magic, offset, size, [crc32, unk1, unk2, unk3, unk4]]


def ARDF_parse_TOC(fd, toc_hdr_size, toc_hdr, toc_entry_func):
    magic_bytes = struct.unpack("<4s", toc_hdr[0:4])[0].decode("ascii")
    unk1 = struct.unpack("<L", toc_hdr[4:8])[0]
    toc_size = struct.unpack("<L", toc_hdr[8:12])[0]
    unk3 = struct.unpack("<L", toc_hdr[12:16])[0]
    toc_entry_count = struct.unpack("<L", toc_hdr[16:20])[0]
    toc_entry_size = struct.unpack("<L", toc_hdr[20:24])[0]

    toc_entries = []
    for i in range(toc_entry_count):
        try:
            entry = toc_entry_func(fd, toc_entry_size)
            if entry != -1:
                toc_entries.append(entry)
        except:
            pass
    return toc_entries


def ARDF_parse_FTOC(fd, ftoc_hdr_size, ftoc_hdr):
    return ARDF_parse_TOC(fd, ftoc_hdr_size, ftoc_hdr, ARDF_get_FTOC_entry)


def ARDF_parse_TTOC(fd, ttoc_hdr_size, ttoc_hdr):
    return ARDF_parse_TOC(fd, ttoc_hdr_size, ttoc_hdr, ARDF_get_TTOC_entry)


def ARDF_load_file(datafile):
    fd = open(datafile, "rb")

    fd.seek(0, os.SEEK_END)
    ardf_file_size = fd.tell()
    fd.seek(0)

    ardf_ftoc = []
    ardf_ttoc = []

    # Reading header
    size, data = ARDF_get_next_element(fd)

    try:
        magic_bytes = struct.unpack("<4s", data[0:4])[0].decode("ascii")
        unknown = struct.unpack("<L", data[4:8])[0]
    except:
        debug_print("Error: load_ardf - header cannot be parsed!")
        return -1

    if magic_bytes != 'ARDF':
        debug_print("Error: load_ardf - header magic bytes do not match!")
        return -1
    else:
        debug_print("Info: load_ardf - found ARDF header.")

    if unknown != 0:
        debug_print("Warning: load_ardf - header does not end with 0... Continuing but something might be wrong!")

    while True:
        try:
            if fd.tell() == ardf_file_size:
                debug_print("Info: File parsed successfully!")
                break
            size, data = ARDF_get_next_element(fd)

            magic_bytes = struct.unpack("<4s", data[0:4])[0].decode("ascii")

            if magic_bytes == 'FTOC':
                # debug_print("Warning: load_ardf - Found FTOC, parsing.")
                ardf_ftoc = ARDF_parse_FTOC(fd, size, data)
            elif magic_bytes == 'TTOC':
                # debug_print("Warning: load_ardf - Found TTOC, parsing.")
                ardf_ttoc = ARDF_parse_TTOC(fd, size, data)
            else:
                break

        except:
            if ardf_ftoc == [] or ardf_ttoc == []:
                debug_print("Error: Could not find FTOC and TTOC!")
            break
    return ardf_ftoc, ardf_ttoc, fd


def ARDF_get_TEXT_from_TOFF(fd, toff_entry):
    toff_entry_offset = toff_entry[1]
    fd.seek(toff_entry_offset)
    toff_entry = ARDF_get_next_element(fd)

    toff_entry_size = toff_entry[0]
    toff_entry_data = toff_entry[1]

    magic_bytes = struct.unpack("<4s", toff_entry_data[0:4])[0].decode("ascii")
    unk1 = struct.unpack("<L", toff_entry_data[4:8])[0]
    unk2 = struct.unpack("<L", toff_entry_data[8:12])[0]
    text_size = struct.unpack("<L", toff_entry_data[12:16])[0]
    text_data = toff_entry_data[16:16 + text_size]

    return text_data


def ARDF_get_VOLM(fd, volm_entry):
    volm_entry_offset = volm_entry[1]
    fd.seek(volm_entry_offset)
    volm_entry = ARDF_get_next_element(fd)
    volm_ftoc = ARDF_parse_FTOC(fd, volm_entry[0], volm_entry[1])
    volm_ttoc_entry = ARDF_get_next_element(fd)
    volm_ttoc = ARDF_parse_TTOC(fd, volm_ttoc_entry[0], volm_ttoc_entry[1])

    vdef_entry = ARDF_get_next_element(fd)[1]
    magic_bytes = struct.unpack("<4s", vdef_entry[0:4])[0].decode("ascii")
    unk1 = struct.unpack("<L", vdef_entry[4:8])[0]
    x_pixcount = struct.unpack("<L", vdef_entry[8:12])[0]
    y_pixcount = struct.unpack("<L", vdef_entry[12:16])[0]
    unk2 = struct.unpack("<d", vdef_entry[16:24])[0]
    unk3 = struct.unpack("<d", vdef_entry[24:32])[0]
    unk4 = struct.unpack("<d", vdef_entry[32:40])[0]
    x_step = struct.unpack("<d", vdef_entry[40:48])[0]
    y_step = struct.unpack("<d", vdef_entry[48:56])[0]
    z_step = struct.unpack("<d", vdef_entry[56:64])[0]
    x_unit = struct.unpack("<32s", vdef_entry[64:96])[0].decode("ascii")
    y_unit = struct.unpack("<32s", vdef_entry[96:128])[0].decode("ascii")
    z_unit = struct.unpack("<32s", vdef_entry[128:160])[0].decode("ascii")
    section_names = struct.unpack("<32s", vdef_entry[160:192])[0].decode("ascii")
    section_number = struct.unpack("<L", vdef_entry[192:196])[0]

    volm_vdef = [section_names, section_number, x_pixcount, y_pixcount, x_step, y_step, z_step, x_unit, y_unit, z_unit]

    x_channel_names = []
    data_channels_names = []
    data_channels_units = []
    voff_data = []
    vset_data = []
    vnam_data = []
    vdat_data = []
    xdat_data = []

    while True:
        try:
            element_data = ARDF_get_next_element(fd)[1]
        except:
            break
        magic_bytes = struct.unpack("<4s", element_data[0:4])[0].decode("ascii")

        if magic_bytes == 'VCHN':
            unk1 = struct.unpack("<L", element_data[4:8])[0]
            channame = struct.unpack("<32s", element_data[8:40])[0]
            chanunit = struct.unpack("<32s", element_data[40:72])[0]
            data_channels_names.append(channame)
            data_channels_units.append(chanunit)
        elif magic_bytes == 'XDEF':
            unk1 = struct.unpack("<L", element_data[4:8])[0]
            unk2 = struct.unpack("<L", element_data[8:12])[0]
            x_length = struct.unpack("<L", element_data[12:16])[0]
            x_channel_names = struct.unpack("<" + str(int(x_length)) + "s", element_data[16:16 + int(x_length)])[0]
        elif magic_bytes == 'VTOC':
            pass
        elif magic_bytes == 'VOFF':
            unk1 = struct.unpack("<L", element_data[4:8])[0]
            unk2 = struct.unpack("<L", element_data[8:12])[0]
            unk3 = struct.unpack("<L", element_data[12:16])[0]
            unk4 = struct.unpack("<L", element_data[16:20])[0]
            unk5 = struct.unpack("<L", element_data[20:24])[0]
            unk6 = struct.unpack("<L", element_data[24:28])[0]
            voff_data.append([unk1, unk2, unk3, unk4, unk5, unk6])
        elif magic_bytes == 'VSET':
            unk1 = struct.unpack("<L", element_data[4:8])[0]
            num_point = struct.unpack("<L", element_data[8:12])[0]
            y_coord = struct.unpack("<L", element_data[12:16])[0]
            x_coord = struct.unpack("<L", element_data[16:20])[0]
            unk5 = struct.unpack("<L", element_data[20:24])[0]
            prev_offset = struct.unpack("<L", element_data[24:28])[0]
            next_offset = struct.unpack("<L", element_data[32:36])[0]
            unk8 = struct.unpack("<L", element_data[36:40])[0]
            vset_data.append([num_point, y_coord, x_coord, unk5, prev_offset, next_offset, unk8])
        elif magic_bytes == 'VNAM':
            unk1 = struct.unpack("<L", element_data[4:8])[0]
            num_point = struct.unpack("<L", element_data[8:12])[0]
            y_coord = struct.unpack("<L", element_data[12:16])[0]
            x_coord = struct.unpack("<L", element_data[16:20])[0]
            name_length = struct.unpack("<L", element_data[20:24])[0]
            meas_name = struct.unpack("<" + str(int(name_length)) + "s", element_data[24:24 + int(name_length)])[0]
            vnam_data.append([num_point, y_coord, x_coord, name_length, meas_name])
        elif magic_bytes == 'VDAT':
            unk1 = struct.unpack("<L", element_data[4:8])[0]
            num_point = struct.unpack("<L", element_data[8:12])[0]
            y_coord = struct.unpack("<L", element_data[12:16])[0]
            x_coord = struct.unpack("<L", element_data[16:20])[0]
            data_length = struct.unpack("<L", element_data[20:24])[0]
            channel_index = struct.unpack("<L", element_data[24:28])[0]
            cut1 = struct.unpack("<L", element_data[32:36])[0]
            cut2 = struct.unpack("<L", element_data[36:40])[0]
            cut3 = struct.unpack("<L", element_data[40:44])[0]
            cut4 = struct.unpack("<L", element_data[44:48])[0]
            meas_data = struct.unpack("<" + str(int(data_length)) + "f", element_data[48:48 + 4 * int(data_length)])
            vdat_data.append(
                [num_point, y_coord, x_coord, data_length, channel_index, cut1, cut2, cut3, cut4, meas_data])
        elif magic_bytes == 'XDAT':
            unk1 = struct.unpack("<L", element_data[4:8])[0]
            num_point = struct.unpack("<L", element_data[8:12])[0]
            y_coord = struct.unpack("<L", element_data[12:16])[0]
            x_coord = struct.unpack("<L", element_data[16:20])[0]
            num_xdata_points = struct.unpack("<L", element_data[20:24])[0]

            xdata_values = []
            start = 24
            for i in range(num_xdata_points):
                end = start + 8
                xdata_values.append(struct.unpack("<d", element_data[start:end])[0])
                start = end
            xdat_data.append([num_point, y_coord, x_coord, num_xdata_points, xdata_values])
        elif magic_bytes == 'MLOV':
            pass
        else:
            continue

    return [volm_ftoc, volm_ttoc, volm_vdef, x_channel_names, data_channels_names, data_channels_units, voff_data,
            vset_data, vnam_data, vdat_data, xdat_data]


def ARDF_get_IMAG(fd, imag_entry):
    imag_entry_offset = imag_entry[1]
    fd.seek(imag_entry_offset)
    imag_entry = ARDF_get_next_element(fd)
    imag_ftoc = ARDF_parse_FTOC(fd, imag_entry[0], imag_entry[1])
    imag_ttoc_entry = ARDF_get_next_element(fd)
    imag_ttoc = ARDF_parse_TTOC(fd, imag_ttoc_entry[0], imag_ttoc_entry[1])

    idef_entry = ARDF_get_next_element(fd)[1]

    magic_bytes = struct.unpack("<4s", idef_entry[0:4])[0].decode("ascii")
    unk1 = struct.unpack("<L", idef_entry[4:8])[0]
    x_pixcount = struct.unpack("<L", idef_entry[8:12])[0]
    y_pixcount = struct.unpack("<L", idef_entry[12:16])[0]
    unk2 = struct.unpack("<d", idef_entry[16:24])[0]
    unk3 = struct.unpack("<d", idef_entry[24:32])[0]
    x_step = struct.unpack("<d", idef_entry[32:40])[0]
    y_step = struct.unpack("<d", idef_entry[40:48])[0]
    x_unit = struct.unpack("<32s", idef_entry[48:80])[0].decode("ascii")
    y_unit = struct.unpack("<32s", idef_entry[80:112])[0].decode("ascii")
    chan_name = struct.unpack("<32s", idef_entry[112:144])[0].decode("ascii")
    z_unit = struct.unpack("<32s", idef_entry[144:176])[0].decode("ascii")

    imag_idef = [chan_name, x_pixcount, y_pixcount, x_step, y_step, x_unit, y_unit, z_unit]

    ibox_entry = ARDF_get_next_element(fd)[1]
    magic_bytes = struct.unpack("<4s", ibox_entry[0:4])[0].decode("ascii")
    unk1 = struct.unpack("<L", ibox_entry[4:8])[0]
    total_size = struct.unpack("<L", ibox_entry[8:12])[0]
    unk2 = struct.unpack("<L", ibox_entry[12:16])[0]
    num_lines = struct.unpack("<L", ibox_entry[16:20])[0]
    line_size = struct.unpack("<L", ibox_entry[20:24])[0]

    imag_ibox = [num_lines]

    imag_data = []

    for i in range(num_lines):
        idat_entry = ARDF_get_next_element(fd)[1]
        magic_bytes = struct.unpack("<4s", idat_entry[0:4])[0].decode("ascii")
        unk1 = struct.unpack("<L", idat_entry[4:8])[0]
        line_data = struct.unpack("<" + str(x_pixcount) + "f", idat_entry[8:8 + 4 * x_pixcount])
        imag_data.append(line_data)

    return [imag_ftoc, imag_ttoc, imag_idef, imag_ibox, imag_data]


def ARDF_parse_THMB(fd, thmb_entry):
    thmb_entry_offset = thmb_entry[1]
    fd.seek(thmb_entry_offset)
    thmb_entry = ARDF_get_next_element(fd)[1]

    magic_bytes = struct.unpack("<4s", thmb_entry[0:4])[0].decode("ascii")
    unk1 = struct.unpack("<L", thmb_entry[4:8])[0]
    x_size = struct.unpack("<L", thmb_entry[8:12])[0]
    y_size = struct.unpack("<L", thmb_entry[12:16])[0]
    bit_depth = struct.unpack("<L", thmb_entry[16:20])[0]
    unk2 = struct.unpack("<L", thmb_entry[20:24])[0]
    nbytes = int(x_size * y_size * (bit_depth / 8))
    bin_data = struct.unpack("<" + str(nbytes) + "s", thmb_entry[24:24 + nbytes])[0]

    return [x_size, y_size, bin_data]


def ardf2hdf5(filename, filepath=None):
    gc.enable()
    ftoc, ttoc, fd = ARDF_load_file(filename)

    # print("======================================")
    # print("Going through global FTOC:")
    for f in ftoc:
        # print(f)
        if f[0] == 'IMAG':
            # print("Found IMAG header, parsing:")
            imag_ftoc, imag_ttoc, imag_idef, imag_ibox, imag_data = ARDF_get_IMAG(fd, f)
            for imag_f in imag_ftoc:
                if imag_f[0] == 'NEXT':
                    # print("--> Ignoring NEXT header for now.")
                    pass
                elif imag_f[0] == 'THMB':
                    # print("--> Parsing THMB (thumbnail):")
                    [x, y, d] = ARDF_parse_THMB(fd, imag_f)
                    d_img = np.array(bytearray(d), dtype=np.byte).reshape((x, y))
                else:
                    print("Found '" + str(imag_f[0]) + "' header. What is it?")
            for imag_t in imag_ttoc:
                if imag_t[0] == 'TOFF':
                    # print("--> Found TOFF, parsing TEXT.")
                    pass
                else:
                    print("Found '" + str(imag_t[0]) + "' header. What is it?")
        elif f[0] == 'VOLM':
            # print("Found VOLM header, parsing:")
            volm_ftoc, volm_ttoc, volm_vdef, x_channel_names, data_channels_names, data_channels_units, voff_data, vset_data, vnam_data, vdat_data, xdat_data = ARDF_get_VOLM(
                fd, f)

            with h5py.File(filename.split('.')[0] + ".hdf5", "w") as f:
                metadatagrp = f.create_group("metadata")
                f.create_group("process")

                if filepath is not None:
                    datagrp = f.create_group("datasets/" + filepath.split('.')[0])
                    datagrp.attrs.__setattr__('type', filepath.split('.')[-1])
                else:
                    datagrp = f.create_group("datasets/" + filename.split('.')[0])
                    datagrp.attrs.__setattr__('type', filename.split('.')[-1])

                label_list = list(map(lambda x: x.decode('utf-8').replace('\x00', ''), data_channels_names))
                y_size = volm_vdef[2]
                x_size = volm_vdef[3]
                z_size = vdat_data[0][6] - vdat_data[0][5]
                numchans = len(label_list)
                data_matrix = np.ndarray(shape=(numchans, y_size, x_size, z_size))
                for vd in vdat_data:
                    try:
                        ypos = vd[1]
                        xpos = vd[2]
                        chanindex = vd[4]
                        data = vd[9][vd[5]:vd[6]]
                        data_matrix[chanindex, ypos, xpos, :] = data[:]
                    except:
                        pass

                for i, k in enumerate(label_list):

                    datagrp.create_dataset(k, data=data_matrix[i, :, :, :])
                    try:
                        datagrp[k].attrs['unit'] = data_channels_units[i].decode('utf-8').replace('\x00', '')
                    except:
                        if b'\xb0' in data_channels_units[i]:
                            datagrp[k].attrs['unit'] = 'deg'
                        else:
                            datagrp[k].attrs['unit'] = 'unknown'
                    datagrp[label_list[i]].attrs['name'] = k
                    datagrp[label_list[i]].attrs['shape'] = data_matrix.shape
                    # datagrp[label_list[i]].attrs['size'] = (fastsize,slowsize)
                    # datagrp[label_list[i]].attrs['offset'] = (xoffset,yoffset)

        else:
            print("Found '" + str(f[0]) + "' header. What is it?")
    # print("======================================")
    # print("Going through global TTOC:")
    for t in ttoc:
        if t[0] == 'TOFF':
            meta = ARDF_get_TEXT_from_TOFF(fd, t).decode('windows-1252')
            metalist = meta.split('\r')
            metalist = [i.encode('utf8') for i in metalist]
        else:
            print("Found '" + str(t[0]) + "' header. What is it?")

        with h5py.File(filename.split('.')[0] + ".hdf5", "a") as f:
            if filepath is not None:
                f["metadata"].create_dataset(filepath.split('.')[0], data=metalist)
            else:
                f["metadata"].create_dataset(filename.split('.')[0], data=metalist)

    gc.disable()



from igor import binarywave
import h5py
from numpy import flipud


# ==========================================
# IBW conversion

def correct_label(label):
    label = [x for x in label if x]  # Remove the empty lists
    label = label[0]  # Remove the unnecessary inception

    corrected_label = []

    for i in label:
        i = i.decode('UTF-8')
        if len(i) == 0:  # Skip empty channel names
            pass
        else:  # Correct the duplicate letters
            if 'Trace' in i:
                i = i.split('Trace')[0]
                corrected_label.append(i + 'Trace')
            elif 'Retrace' in i:
                i = i.split('Retrace')[0]
                corrected_label.append(i + 'Retrace')
            else:
                corrected_label.append(i)
    corrected_label = [x.encode() for x in corrected_label]
    return corrected_label


def ibw2hdf5(filename, filepath=None):
    tmpdata = binarywave.load(filename)['wave']
    note = tmpdata['note']
    label_list = correct_label(tmpdata['labels'])

    fastsize = float(str(note).split('FastScanSize:')[-1].split('\\r')[0])
    slowsize = float(str(note).split('SlowScanSize:')[-1].split('\\r')[0])
    xoffset = float(str(note).split('XOffset:')[1].split('\\r')[0])
    yoffset = float(str(note).split('YOffset:')[1].split('\\r')[0])
    with h5py.File(filename.split('.')[0] + ".hdf5", "w") as f:
        f.require_group("process")
        metadatagrp = f.require_group("metadata")
        if filepath is not None:
            metadatagrp.create_dataset(filepath.split('.')[0], data=tmpdata['note'])
            datagrp = f.require_group("datasets/" + filepath.split('.')[0])
            datagrp.attrs.__setattr__('type', filepath.split('.')[-1])

        else:
            metadatagrp.create_dataset(filename.split('.')[0], data=tmpdata['note'])
            datagrp = f.require_group("datasets/" + filename.split('.')[0])
            datagrp.attrs.__setattr__('type', filename.split('.')[-1])

        for i, k in enumerate(label_list):
            if len(np.shape(tmpdata['wData'])) == 2:
                datagrp.create_dataset(k, data=flipud(tmpdata['wData'][:, i].T))
                datagrp[label_list[i]].attrs['shape'] = tmpdata['wData'][:, i].T.shape
                datagrp[label_list[i]].attrs['scale_m_per_px'] = fastsize/tmpdata['wData'][:, i].T.shape[0]
            else:
                datagrp.create_dataset(k, data=flipud(tmpdata['wData'][:, :, i].T))
                datagrp[label_list[i]].attrs['shape'] = tmpdata['wData'][:, :, i].T.shape
                datagrp[label_list[i]].attrs['scale_m_per_px'] = fastsize/tmpdata['wData'][:, :, i].T.shape[0]
            datagrp[label_list[i]].attrs['name'] = k.decode('utf8')
            datagrp[label_list[i]].attrs['size'] = (fastsize, slowsize)
            datagrp[label_list[i]].attrs['offset'] = (xoffset, yoffset)
            datagrp[label_list[i]].attrs['path'] = ("datasets/" + filename.split('.')[0]+"/"+str(k).split('\'')[1])


            if "Phase" in str(k):
                datagrp[label_list[i]].attrs['unit'] = ('m', 'm', 'deg')
            elif "Amplitude" in str(k):
                datagrp[label_list[i]].attrs['unit'] = ('m', 'm', 'V')
            elif "Height" in str(k):
                datagrp[label_list[i]].attrs['unit'] = ('m', 'm', 'm')
            else:
                datagrp[label_list[i]].attrs['unit'] = ('m', 'm', 'unknown')
        # f.create_dataset("channelsdata/pxs", data=sizes)

    print('file successfully converted')



try:
    import xrdtools
    xrdtools_bool = True
except ImportError:
    xrdtools_bool = False
import h5py
import re
import numpy as np


# ==========================================
# XRDML conversion

def xrdml2hdf5(filename, filepath=None):
    if not xrdtools_bool:
        print('Please download the xrdtools package if you want to use this function')
        return

    with open(filename, 'r') as f:
        contents = f.read()
        
    with h5py.File(filename.split('.')[0] + ".hdf5", "w") as f:
        param_names = ['counts', '2theta', 'omega', 'phi', 'chi', 'x', 'y']
        param_headers = ['',
                         '<positions axis="2Theta" unit="deg">',
                         '<positions axis="Omega" unit="deg">',
                         '<positions axis="Phi" unit="deg">',
                         '<positions axis="Chi" unit="deg">',
                         '<positions axis="X" unit="mm">',
                         '<positions axis="Z" unit="mm">']

        scans = contents.split('<dataPoints>')[1:]

        metadatagrp = f.create_group("metadata")
        f.create_group("process")

        if filepath is not None:
            metadatagrp.create_dataset(filepath.split('.')[0], data=contents)
            datagrp = f.create_group("datasets/" + filepath.split('.')[0])
            datagrp.attrs.__setattr__('type', filepath.split('.')[-1])

        else:
            metadatagrp.create_dataset(filename.split('.')[0], data=contents)
            datagrp = f.create_group("datasets/" + filename.split('.')[0])
            datagrp.attrs.__setattr__('type', filename.split('.')[-1])

        for i in range(len(param_names)):
            name = param_names[i]
            all_data = []
            for scan in scans:
                if i == 0:
                    counts = scan.split('"counts">')[-1].split('</')[0].split()
                    val_list = list(map(float, counts))
                    step_num = len(val_list)
                if i != 0:
                    data_range = scan.split(param_headers[i])[1].split('</positions>')[0]
                    if len(data_range.split()) == 1:
                        val = float(re.findall(r"[-+]?\d*\.\d+|\d+", data_range.split()[0])[0])
                        val_list = np.full(step_num, val)
                    elif len(data_range.split()) == 2:
                        first_val = float(re.findall(r"[-+]?\d*\.\d+|\d+", data_range.split()[0])[0])
                        last_val = float(re.findall(r"[-+]?\d*\.\d+|\d+", data_range.split()[1])[0])
                        val_list = np.arange(first_val, last_val, (last_val - first_val) / step_num)
                all_data.append(val_list)
            all_data = np.array(all_data).T
            datagrp.create_dataset(name, data=all_data)
            datagrp[name].attrs['shape'] = np.shape(all_data)
            datagrp[name].attrs['name'] = 'counts'
            datagrp[name].attrs['size'] = 0
            datagrp[name].attrs['offset'] = 0
            datagrp[name].attrs['unit'] = 'au'
            if name == 'x' or name == 'y' or name == 'z':
                datagrp[name].attrs['unit'] = 'mm'
            elif name == 'counts':
                datagrp[name].attrs['unit'] = 'au'
            else:
                datagrp[name].attrs['unit'] = 'deg'

    print("File successfully converted")



import setuptools
import pathlib

here = pathlib.Path(__file__).parent.resolve()

setuptools.setup(
    version_config={
        "template": "{tag}",
        "dev_template": "{tag}.dev{ccount}",
        "dirty_template": "{tag}.dev{ccount}",
        "starting_version": "0.0.1",
        "version_callback": None,
        "version_file": None,
        "count_commits_from_version_file": True
    },
    setup_requires=['setuptools-git-versioning'],
    name="hystorian",
    author="Loïc Musy <loic.musy@unige.ch>, Ralph Bulanadi <ralph.bulanadi@unige.ch>",
    author_email="loic.musy@unige.ch",
    description="a generic materials science data analysis Python package built with processing traceability, reproducibility, and archival ability at its core.",
    long_description=(here / 'README.md').read_text(encoding='utf-8'),
    long_description_content_type='text/markdown',
    url='https://gitlab.unige.ch/paruch-group/hystorian',
    packages=['hystorian', 'hystorian.io', 'hystorian.processing'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'h5py>2,<3',
        'numpy>=1.18',
        'igor',
        'pillow>7',
        'scipy',
        'scikit-image',
        'opencv'
    ],
    python_requires='>=3.6',
)
